import copy

from loqusdb.plugins import MongoAdapter
from loqusdb.build_models import build_variant


class TestInsertVariant:

    def test_insert_one_variant(self, mongo_adapter, simplest_variant):
        """Test to insert one variant"""

        mongo_adapter.add_variant(simplest_variant)

        db = mongo_adapter.db

        #Get the variant without the adapter
        mongo_variant = db.variant.find_one()

        assert mongo_variant['_id'] == 'test'
        assert mongo_variant['observations'] == 1
        assert mongo_variant['homozygote'] == 0

    def test_insert_one_variant_twice(self, mongo_adapter, simplest_variant):
        """Test to insert one variant"""

        mongo_adapter.add_variant(simplest_variant)
        mongo_adapter.add_variant(simplest_variant)

        db = mongo_adapter.db

        mongo_variant = db.variant.find_one()

        assert mongo_variant['_id'] == 'test'
        assert mongo_variant['observations'] == 2
        assert mongo_variant.get('homozygote',0) == 0

    def test_insert_hom_variant(self, real_mongo_adapter, homozygous_variant):
        """Test to insert a homozygote variant"""
        mongo_adapter = real_mongo_adapter
        mongo_adapter.add_variant(homozygous_variant)

        db = mongo_adapter.db

        mongo_variant = db.variant.find_one()
        assert mongo_variant['_id'] == 'test'
        assert mongo_variant['observations'] == 1
        assert mongo_variant.get('homozygote', 0) == 1
        assert mongo_variant['families'] == ['1']

    def test_insert_many(self, mongo_adapter, simplest_variant):
        """Test to insert a homozygote variant"""

        for i in range(10000):
            mongo_adapter.add_variant(simplest_variant)

        db = mongo_adapter.db

        mongo_variant = db.variant.find_one()
        assert mongo_variant['_id'] == 'test'
        assert mongo_variant['observations'] == 10000
        assert mongo_variant.get('homozygote', 0) == 0


class TestGetVariant:
    def test_get_variant(self, mongo_client, mongo_adapter, simplest_variant):
        """Test to insert one variant"""

        #Insert without adapter
        db = mongo_client['test']
        db.variant.insert_one(simplest_variant)

        mongo_variant = mongo_adapter.get_variant(simplest_variant)
        assert mongo_variant['_id'] == 'test'

    def test_get_none(self, mongo_adapter, simplest_variant):
        """Test to get non existing variant"""

        mongo_variant = mongo_adapter.get_variant(simplest_variant)

        assert mongo_variant == None

class TestBulkOperations:

    def test_insert_one_variant(self, real_mongo_adapter, simplest_variant):
        """Test to insert one variant with bulk insert"""

        mongo_adapter = real_mongo_adapter
        variants = [simplest_variant]

        mongo_adapter.add_variants(variants)
        db = mongo_adapter.db
        mongo_variant = db.variant.find_one()

        assert mongo_variant['_id'] == 'test'
        assert mongo_variant['observations'] == 1
        assert mongo_variant['homozygote'] == 0

    def test_insert_two_variants(self, real_mongo_adapter):
        """Test to insert two variants with bulk"""

        adapter = real_mongo_adapter
        db = adapter.db

        variants = []
        variants.append({
            '_id': 'test',
            'homozygote': 0
        })
        variants.append({
            '_id': 'test_1',
            'homozygote': 1
        })


        adapter.add_variants(variants)

        first_variant = db.variant.find_one({'_id': 'test'})
        second_variant = db.variant.find_one({'_id': 'test_1'})

        assert first_variant['_id'] == 'test'
        assert first_variant['observations'] == 1
        assert first_variant.get('homozygote',0) == 0

        assert second_variant['_id'] == 'test_1'
        assert second_variant['observations'] == 1
        assert second_variant.get('homozygote',0) == 1

    def test_insert_many(self, real_mongo_adapter):

        adapter = real_mongo_adapter
        db = adapter.db

        variants = ({'_id': 'test','homozygote': 0} for i in range(20000))

        adapter.add_variants(variants)

        mongo_variant = db.variant.find_one()

        assert mongo_variant['_id'] == 'test'
        assert mongo_variant['observations'] == 20000
        assert mongo_variant.get('homozygote', 0) == 0

class TestRemoveVariant:

    def test_remove_one_variant(self, mongo_adapter):
        """Test to update one variant"""

        db = mongo_adapter.db

        variant = {
            '_id': 'test',
            'observations': 1
        }

        db.variant.insert_one(variant)

        mongo_adapter.delete_variant(variant)

        assert db.variant.find_one() == None

    def test_downcount_one_variant(self, mongo_adapter):
        """Test to update one variant"""

        db = mongo_adapter.db

        insert_variant = {
            '_id': 'test',
            'families':['1', '2'],
            'observations': 2
        }

        db.variant.insert_one(insert_variant)

        variant = {
            '_id': 'test',
            'case_id':'1'
        }

        mongo_adapter.delete_variant(variant)

        mongo_variant = db.variant.find_one()

        assert mongo_variant['observations'] == 1
        assert mongo_variant['families'] == ['2']

    def test_remove_non_existing(self, mongo_adapter, simplest_variant):

        db = mongo_adapter.db

        mongo_adapter.delete_variant(simplest_variant)

        assert db.variant.find_one() == None

class TestRemoveSV:

    def test_remove_one_SV(self, mongo_adapter, del_variant, case_obj):

        # GIVEN a database poulated with one SV
        db = mongo_adapter.db
        formated_variant = build_variant(del_variant, case_obj=case_obj, case_id=case_obj['case_id'])
        mongo_adapter.add_structural_variant(formated_variant)
        mongo_SV = db.structural_variant.find_one()
        mongo_identity = db.identity.find_one()
        assert mongo_SV is not None
        assert mongo_identity is not None
        # WHEN deleting SV
        mongo_adapter.delete_structural_variant(formated_variant)

        # THEN there should be no remaining SVs in the database
        mongo_SV = db.structural_variant.find_one()
        mongo_identity = db.indentity.find_one()
        assert mongo_SV is None
        assert mongo_identity is None

    def test_remove_one_of_two_SV(self, mongo_adapter, duptandem_variant, case_obj):

        # GIVEN a database poulated with one SV
        db = mongo_adapter.db
        formated_variant = build_variant(duptandem_variant, case_obj=case_obj, case_id=case_obj['case_id'])
        mongo_adapter.add_structural_variant(formated_variant)

        # Add second of same variant, changing the start and end position slightly
        formated_variant_ = copy.deepcopy(formated_variant)
        formated_variant_['pos'] = formated_variant_['pos']+2
        formated_variant_['end'] = formated_variant_['end']-1
        formated_variant_['case_id'] = 'case_2'
        mongo_adapter.add_structural_variant(formated_variant_)

        # This should correspond to one structural variant document
        mongo_svs = list(db.structural_variant.find())
        assert len(mongo_svs) == 1
        mongo_sv = mongo_svs[0]
        assert mongo_sv['pos_sum'] == formated_variant['pos'] + formated_variant_['pos']
        # And two identity documents
        mongo_identities = list(db.identity.find())
        assert len(mongo_identities) == 2

        # WHEN deleting the variant from the first case
        mongo_adapter.delete_structural_variant(formated_variant)

        # THEN the SV document should have the pos_sum equal to the pos of the
        # SV from the second case
        mongo_svs = list(db.structural_variant.find())
        assert len(mongo_svs) == 1
        mongo_sv = mongo_svs[0]
        assert mongo_sv['pos_sum'] == formated_variant_['pos']
        # And one identity documents
        mongo_identities = list(db.identity.find())
        assert len(mongo_identities) == 1

class TestHelperMethods:

    def test_update_sv_metrics(self, mongo_adapter):
        # GIVEN a mongo adapter

        # WHEN cluster_len > 10000
        cluster_len, interval_size = mongo_adapter._update_sv_metrics(sv_type='INV',
                                                                      pos_mean=10000,
                                                                      end_mean=30000,
                                                                      max_window=3000)
        # THEN interval_size should be the cluster_len divided by 10
        assert cluster_len == 20000
        assert interval_size == round(20000/10, -2)

        # WHEN cluster_len <10000
        cluster_len, interval_size = mongo_adapter._update_sv_metrics(sv_type='DUP',
                                                                      pos_mean=10000,
                                                                      end_mean=15000,
                                                                      max_window=3000)

        # THEN interval_size should be cluster_len divided by 5
        assert cluster_len == 5000
        assert interval_size == round(5000/5, -2)

        # WHEN interval_size < 1000
        cluster_len, interval_size = mongo_adapter._update_sv_metrics(sv_type='DEL',
                                                                      pos_mean=10000,
                                                                      end_mean=10500,
                                                                      max_window=3000)

        # THEN interval_size should be cluster_len divided by 2
        assert cluster_len == 500
        assert interval_size == round(500/2, -2)

        # WHEN interval size is > max_window
        cluster_len, interval_size = mongo_adapter._update_sv_metrics(sv_type='INV',
                                                                      pos_mean=100000,
                                                                      end_mean=200000,
                                                                      max_window=3000)
        # THEN interval size should be set to max_window
        assert cluster_len == 100000
        assert interval_size == 3000

        # WHEN sv_type == BND
        cluster_len, interval_size = mongo_adapter._update_sv_metrics(sv_type='BND',
                                                                      pos_mean=1000,
                                                                      end_mean=2000,
                                                                      max_window=3000)
        # THEN cluster_len should be 10e10 and interval_size 2*max window
        assert cluster_len == 10e10
        assert interval_size == 2*3000



import logging

from pprint import pprint as pp

from loqusdb.plugins import BaseVariantMixin
from .structural_variant import SVMixin

from pymongo import (ASCENDING, DESCENDING, UpdateOne, DeleteOne)

LOG = logging.getLogger(__name__)

class VariantMixin(BaseVariantMixin, SVMixin):

    def _get_update(self, variant):
        """Convert a variant to a proper update

        Args:
            variant(dict)

        Returns:
            update(dict)
        """
        update = {
                '$inc': {
                    'homozygote': variant.get('homozygote', 0),
                    'hemizygote': variant.get('hemizygote', 0),
                    'observations': 1
                },
                '$set': {
                    'chrom': variant.get('chrom'),
                    'start': variant.get('pos'),
                    'end': variant.get('end'),
                    'ref': variant.get('ref'),
                    'alt': variant.get('alt'),
                }
             }
        if variant.get('case_id'):
            update['$push'] = {
                                'families': {
                                '$each': [variant.get('case_id')],
                                '$slice': -50
                                }
                            }
        return update

    def _get_update_delete(self, variant):
        update = {
            '$inc': {
                'homozygote': - variant.get('homozygote', 0),
                'hemizygote': - variant.get('hemizygote', 0),
                'observations': -1
                },
            '$set': {
                'chrom': variant.get('chrom'),
                'start': variant.get('pos'),
                'end': variant.get('end'),
                'ref': variant.get('ref'),
                'alt': variant.get('alt'),
                }
            }
        if variant.get('case_id'):
            update['$pull'] = {'families': variant.get('case_id')}

        return update

    def add_variant(self, variant):
        """Add a variant to the variant collection

            If the variant exists we update the count else we insert a new variant object.

            Args:
                variant (dict): A variant dictionary

        """
        LOG.debug("Upserting variant: {0}".format(variant.get('_id')))

        update = self._get_update(variant)

        message = self.db.variant.update_one(
            {'_id': variant['_id']},
            update,
            upsert=True
        )
        if message.modified_count == 1:
            LOG.debug("Variant %s was updated", variant.get('_id'))
        else:
            LOG.debug("Variant was added to database for first time")
        return

    def add_variants(self, variants):
        """Add a bulk of variants

        This could be used for faster inserts

        Args:
            variants(iterable(dict))

        """

        operations = []
        nr_inserted = 0
        for i,variant in enumerate(variants, 1):
            # We need to check if there was any information returned
            # The variant could be excluded based on low gq or if no individiual was called
            # in the particular case
            if not variant:
                continue
            nr_inserted += 1
            update = self._get_update(variant)
            operations.append(
                UpdateOne(
                    {'_id': variant['_id']},
                    update,
                    upsert=True
                )
            )
            if i % 10000 == 0:
                self.db.variant.bulk_write(operations, ordered=False)
                operations = []

        if len(operations) > 0:
            self.db.variant.bulk_write(operations, ordered=False)

        return nr_inserted

    def get_variant(self, variant):
        """Check if a variant exists in the database and return it.
        Variants are searched with a variant id which is a string that consists of
        chrom_pos_ref_alt.
        There is no simple way to create a similar ID for structural variants so this function
        will not work for SVs

        Search the variants with the variant id

        Args:
            variant(dict): A variant dictionary

        Returns:
            variant(dict): A variant dictionary or None
        """
        return self.db.variant.find_one({'_id': variant.get('_id')})

    def search_variants(self, variant_ids):
        """Make a batch search for variants in the database

        Args:
            variant_ids(list(str)): List of variant ids

        Returns:
            res(pymngo.Cursor(variant_obj)): The result
        """
        query = {'_id': {'$in': variant_ids}}

        return self.db.variant.find(query)


    def get_variants(self, chromosome=None, start=None, end=None):
        """Return all variants in the database
        If no region is specified all variants will be returned.

        Args:
            chromosome(str)
            start(int)
            end(int)


        Returns:
            variants(Iterable(Variant))
        """
        query = {}
        if chromosome:
            query['chrom'] = chromosome
        if start:
            query['start'] = {'$lte': end}
            query['end'] = {'$gte': start}
        LOG.info("Find all variants {}".format(query))
        return self.db.variant.find(query).sort([('start', ASCENDING)])
    
    def nr_variants(self, chromosome=None, start=None, end=None):
        """Return nr of variants"""
        
        query = {}
        if chromosome:
            query['chrom'] = chromosome
        if start:
            query['start'] = {'$lte': end}
            query['end'] = {'$gte': start}
        LOG.info("Find all variants {}".format(query))
        return self.db.variant.count_documents(query)
        
        
    def delete_variant(self, variant):
        """Delete observation in database

        This means that we take down the observations variable with one.
        If 'observations' == 1 we remove the variant. If variant was homozygote
        we decrease 'homozygote' with one.
        Also remove the family from array 'families'.

        Args:
            variant (dict): A variant dictionary

        """
        mongo_variant = self.get_variant(variant)

        if mongo_variant:

            if mongo_variant['observations'] == 1:
                LOG.debug("Removing variant {0}".format(
                    mongo_variant.get('_id')
                ))
                message = self.db.variant.delete_one({'_id': variant['_id']})
            else:
                LOG.debug("Decreasing observations for {0}".format(
                    mongo_variant.get('_id')
                ))
                message = self.db.variant.update_one({
                    '_id': mongo_variant['_id']
                    },{
                        '$inc': {
                            'observations': -1,
                            'homozygote': - (variant.get('homozygote', 0)),
                            'hemizygote': - (variant.get('hemizygote', 0)),
                        },
                        '$pull': {
                            'families': variant.get('case_id')
                        }
                    }, upsert=False)
        return


    def delete_variants(self, variants):
        """Delete observations in database

        Given a list of variants, the write operation for each of the variant
        is given as a bulk to mongodb.

        Args:
            variants (list(Variant)): a list of variants
        """

        variant_id_dict = {variant['_id']: variant for variant in variants}
        # Look up all variants at the same time to reduce number of operations
        # done on the database
        query = self.db.variant.find({'_id': {'$in': list(variant_id_dict.keys())}})
        operations = []
        for mongo_variant in query:
            variant = variant_id_dict.get(mongo_variant['_id'])
            if variant is None:
                continue
            if mongo_variant['observations'] == 1:
                operations.append(DeleteOne({'_id': variant['_id']}))
                continue
            update = self._get_update_delete(variant)
            operations.append(
                UpdateOne(
                    {'_id': variant['_id']},
                    update,
                    upsert=False
                )
            )
        # Make the accumulated write operations
        if len(operations) > 0:
            self.db.variant.bulk_write(operations, ordered=False)

    def get_chromosomes(self, sv=False):
        """Return a list of all chromosomes found in database

        Args:
            sv(bool): if sv variants should be choosen

        Returns:
            res(iterable(str)): An iterable with all chromosomes in the database
        """
        if sv:
            res = self.db.structural_variant.distinct('chrom')
        else:
            res = self.db.variant.distinct('chrom')

        return res

    def get_max_position(self, chrom):
        """Get the last position observed on a chromosome in the database

        Args:
            chrom(str)

        Returns:
            end(int): The largest end position found

        """
        res = self.db.variant.find({'chrom':chrom}, {'_id':0, 'end':1}).sort([('end', DESCENDING)]).limit(1)
        end = 0
        for variant in res:
            end = variant['end']
        return end



"""
loqusdb/utils/annotate.py

Functions to annotate variants in a VCF

The methods for annotating SNVs and SVs differs a bit so they will be handeld in separate.

"""

from loqusdb.build_models.variant import (get_variant_id, get_coords)

def annotate_variant(variant, var_obj=None):
    """Annotate a cyvcf variant with observations
    
    Args:
        variant(cyvcf2.variant)
        var_obj(dict)
    
    Returns:
        variant(cyvcf2.variant): Annotated variant
    """
    if var_obj:
    
        variant.INFO['Obs'] = var_obj['observations']
        if var_obj.get('homozygote'):
            variant.INFO['Hom'] = var_obj['homozygote']
        if var_obj.get('hemizygote'):
            variant.INFO['Hem'] = var_obj['hemizygote']
    
    return variant

def annotate_snv(adpter, variant):
    """Annotate an SNV/INDEL variant
    
    Args:
        adapter(loqusdb.plugin.adapter)
        variant(cyvcf2.Variant)
    """
    variant_id = get_variant_id(variant)
    variant_obj = adapter.get_variant(variant={'_id':variant_id})
    
    annotated_variant = annotated_variant(variant, variant_obj)
    return annotated_variant

def annotate_svs(adapter, vcf_obj):
    """Annotate all SV variants in a VCF
    
    Args:
        adapter(loqusdb.plugin.adapter)
        vcf_obj(cyvcf2.VCF)
    
    Yields:
        variant(cyvcf2.Variant)
    """
    for nr_variants, variant in enumerate(vcf_obj, 1):
        variant_info = get_coords(variant)
        match = adapter.get_structural_variant(variant_info)
        if match:
            annotate_variant(variant, match)
        yield variant
    

def annotate_snvs(adapter, vcf_obj):
    """Annotate all variants in a VCF
    
    Args:
        adapter(loqusdb.plugin.adapter)
        vcf_obj(cyvcf2.VCF)
    
    Yields:
        variant(cyvcf2.Variant): Annotated variant
    """
    variants = {}
    
    for nr_variants, variant in enumerate(vcf_obj, 1):
        # Add the variant to current batch
        variants[get_variant_id(variant)] = variant
        # If batch len == 1000 we annotate the batch
        if (nr_variants % 1000) == 0:
            
            for var_obj in adapter.search_variants(list(variants.keys())):
                var_id = var_obj['_id']
                if var_id in variants:
                    annotate_variant(variants[var_id], var_obj)

            for variant_id in variants:
                yield variants[variant_id]
            
            variants = {}
    
    for var_obj in adapter.search_variants(list(variants.keys())):
        var_id = var_obj['_id']
        if var_id in variants:
            annotate_variant(variants[var_id], var_obj)

    for variant_id in variants:
        yield variants[variant_id]



import os
import logging
import click

from pprint import pprint as pp

from datetime import datetime

from loqusdb.exceptions import (VcfError)
from loqusdb.utils.load import load_database
from loqusdb.utils.vcf import (get_file_handle, check_vcf, add_headers)
from loqusdb.utils.annotate import (annotate_snvs, annotate_svs)

from . import base_command

LOG = logging.getLogger(__name__)

@base_command.command('annotate', short_help="Annotate a VCF with observations")
@click.argument('variant-file',
                    type=click.Path(exists=True),
                    metavar='<vcf_file>',
)
@click.option('--sv', is_flag=True)
@click.pass_context
def annotate(ctx, variant_file, sv):
    """Annotate the variants in a VCF

    """
    adapter = ctx.obj['adapter']

    variant_path = os.path.abspath(variant_file)

    expected_type = 'snv'
    if sv:
        expected_type = 'sv'

    if 'sv':
        nr_cases = adapter.nr_cases(sv_cases=True)
    else:
        nr_cases = adapter.nr_cases(snv_cases=True)
    LOG.info("Found {0} {1} cases in database".format(nr_cases, expected_type))

    vcf_obj = get_file_handle(variant_path)
    add_headers(vcf_obj, nr_cases=nr_cases, sv=sv)
    # Print the headers
    for header_line in vcf_obj.raw_header.split('\n'):
        if len(header_line) == 0:
            continue
        click.echo(header_line)

    start_inserting = datetime.now()
    
    if sv:
        annotated_variants = annotate_svs(adapter, vcf_obj)
    else:
        annotated_variants = annotate_snvs(adapter, vcf_obj)
    # try:
    for variant in annotated_variants:
        click.echo(str(variant).rstrip())
    # except (Exception) as error:
    #     LOG.warning(error)
    #     ctx.abort()
    
    



import logging

LOG = logging.getLogger(__name__)

class ProfileVariantMixin():

    def add_profile_variants(self, profile_variants):

        """Add several variants to the profile_variant collection in the
        database

        Args:

            profile_variants(list(models.ProfileVariant))

        """

        results = self.db.profile_variant.insert_many(profile_variants)

        return results

    def profile_variants(self):
        """Get all profile variants from the database

        Returns:
            profile_variants (Iterable(ProfileVariant))
        """

        return self.db.profile_variant.find()



import pytest

from loqusdb.utils.profiling import (get_profiles,
                                     profile_match,
                                     compare_profiles,
                                     check_duplicates)
from loqusdb.utils.load import (load_profile_variants, load_case)
from loqusdb.utils.vcf import check_vcf

from loqusdb.exceptions import ProfileError

def test_get_profiles(real_mongo_adapter, profile_vcf_path, zipped_vcf_path):

    #Load profile variants
    load_profile_variants(real_mongo_adapter, profile_vcf_path)

    vcf_info = check_vcf(zipped_vcf_path)

    #Get profiles from vcf
    profiles = get_profiles(real_mongo_adapter, zipped_vcf_path)

    #Assert that all individuals are included
    assert list(profiles.keys()) == vcf_info['individuals']

    #Assert that profile strings are of same lengths
    for i, individual in enumerate(profiles.keys()):
        if i==0: length = len(profiles[individual])
        assert len(profiles[individual]) == length


def test_profile_match(real_mongo_adapter, profile_vcf_path, profile_list, case_obj):

    #Load profile variants
    load_profile_variants(real_mongo_adapter, profile_vcf_path)

    #Load case having profiles profile_list
    load_case(real_mongo_adapter, case_obj)

    #Get profiles from vcf
    profiles = {'test_individual': profile_list}

    #Assert that error is raised
    with pytest.raises(ProfileError) as error:

        profile_match(real_mongo_adapter, profiles)


def test_check_duplicates(real_mongo_adapter, profile_vcf_path, profile_list, case_obj):

    # Load profile variants
    load_profile_variants(real_mongo_adapter, profile_vcf_path)
    # Load case having profiles profile_list
    load_case(real_mongo_adapter, case_obj)
    # Create profiles dictionary
    profiles = {'test_individual': profile_list}
    # match profiles to the profiles in the database
    match = check_duplicates(real_mongo_adapter, profiles, hard_threshold=0.95)
    # This should match with the sample in the database
    assert match['profile'] == profile_list

    # Change last genotype, now no matches should be found
    profiles = {'test_individual': profile_list[:-1] + ['NN']}
    match = check_duplicates(real_mongo_adapter, profiles, hard_threshold=0.80)
    assert match is None

    # Lower threshold. Now match should be found
    match = check_duplicates(real_mongo_adapter, profiles, hard_threshold=0.75)
    assert match['profile'] == profile_list

def test_compare_profiles():

    assert compare_profiles(['AA', 'CC'], ['AA', 'CC']) == 1.0
    assert compare_profiles(['AA', 'CC'], ['GG', 'CC']) == 0.5
    assert compare_profiles(['AC', 'CG'], ['TC', 'CG']) == 0.5
    assert compare_profiles(['AC', 'GT'], ['AA', 'GG']) == 0.0



# -*- coding: utf-8 -*-
import logging
import json

import click

from pprint import pprint as pp

from . import base_command

LOG = logging.getLogger(__name__)

@base_command.command('cases', short_help="Display cases in database")
@click.option('-c' ,'--case-id',
                help='Search for case'
)
@click.option('--to-json', is_flag=True)
@click.option('--count', is_flag=True, 
    help="Show number of cases in database")
@click.option('--case-type', '-t', 
    help="Specify the type of cases", 
    type=click.Choice(['snv', 'sv'])
)
@click.pass_context
def cases(ctx, case_id, to_json, count, case_type):
    """Display cases in the database."""

    adapter = ctx.obj['adapter']
    cases = []
    
    if count:
        snv_cases = None
        sv_cases = None
        if case_type == 'snv':
            snv_cases = True
        if case_type == 'sv':
            sv_cases = True
        click.echo(adapter.nr_cases(snv_cases=snv_cases, sv_cases=sv_cases))
        return

    if case_id:
        case_obj = adapter.case({'case_id':case_id})
        if not case_obj:
            LOG.info("Case {0} does not exist in database".format(case_id))
            return
        case_obj['_id'] = str(case_obj['_id'])
        cases.append(case_obj)
    else:
        cases = adapter.cases()
        if cases.count() == 0:
            LOG.info("No cases found in database")
            ctx.abort()

    if to_json:
        click.echo(json.dumps(cases))
        return

    click.echo("#case_id\tvcf_path")

    for case_obj in cases:
        case_id = case_obj.get('case_id')
        vcf_path = case_obj.get('vcf_path', case_obj.get('vcf_sv_path'))
        click.echo("{0}\t{1}".format(case_id, vcf_path))

@base_command.command('variants', short_help="Display variants in database")
@click.option('--variant-id',
                help='Search for a variant'
)
@click.option('-c', '--chromosome',
                help='Search for all variants in a chromosome'
)
@click.option('--end-chromosome',
                help='Search for all variants that ends on chromosome'
)
@click.option('-s', '--start',
                help='Start of region',
                type=int
)
@click.option('-e', '--end',
                help='End of region',
                type=int
)
@click.option('-t', '--variant-type',
                help='Variant type to search for',
                type=click.Choice(['sv', 'snv']),
                default='snv'
)
@click.option('--sv-type',
                help='Type of svs to search for',
)
@click.option('--to-json', is_flag=True)
@click.option('--case-count', is_flag=True, help="If number of cases should be included")
# @click.option('--sort-key',
#                 help='Specify what field to sort on',
# )
@click.pass_context
def variants(ctx, variant_id, chromosome, end_chromosome, start, end, variant_type,
             sv_type, to_json, case_count):
    """Display variants in the database."""
    if sv_type:
        variant_type = 'sv'

    adapter = ctx.obj['adapter']

    if (start or end):
        if not (chromosome and start and end):
            LOG.warning("Regions must be specified with chromosome, start and end")
            return
    
    nr_cases = None
    if case_count:
        snv_cases = None
        sv_cases = None
        if variant_type == 'snv':
            snv_cases = True
        if variant_type == 'sv':
            sv_cases = True
        nr_cases = adapter.nr_cases(snv_cases=snv_cases, sv_cases=sv_cases)

    if variant_id:
        if variant_type == 'sv':
            variant_query = {
                'chrom': chromosome,
                'end_chrom': end_chromosome or chromosome,
                'sv_type': sv_type,
                'pos': start,
                'end': end
            }
            variant = adapter.get_structural_variant(variant_query)
        else:
            variant = adapter.get_variant({'_id':variant_id})
        
        if not variant:
            LOG.info("Variant {0} does not exist in database".format(variant_id))
            variant = {}
        
        if case_count:
            variant['total'] = nr_cases

        if to_json:
            LOG.info("Print in json format")
            if '_id' in variant:
                variant.pop('_id')
            click.echo(json.dumps(variant))
            return

        click.echo(variant)
        return

    if variant_type == 'snv':
        result = adapter.get_variants(
            chromosome=chromosome,
            start=start,
            end=end
        )
    else:
        LOG.info("Search for svs")
        result = adapter.get_sv_variants(
            chromosome=chromosome,
            end_chromosome=end_chromosome,
            sv_type=sv_type,
            pos=start,
            end=end
        )

    if to_json:
        json.dumps(variant)

    i = 0
    for i,variant in enumerate(result,1):
        pp(variant)

    LOG.info("Number of variants found in database: %s", i)

@base_command.command('index', short_help="Add indexes to database")
@click.option('--view',
    is_flag=True,
    help='Only display existing indexes',
)
@click.pass_context
def index(ctx, view):
    """Index the database."""
    adapter = ctx.obj['adapter']
    if view:
        click.echo(adapter.indexes())
        return
    adapter.ensure_indexes()



import logging

from loqusdb.models import ProfileVariant
from .variant import get_variant_id

LOG = logging.getLogger(__name__)

def get_maf(variant):
    """
        Gets the MAF (minor allele frequency) tag from the info field for the
        variant.

        Args:
            variant (cyvcf2.Variant)

        Returns:
            maf (float): Minor allele frequency

    """

    return variant.INFO.get('MAF')


def build_profile_variant(variant):
    """Returns a ProfileVariant object

    Args:
        variant (cyvcf2.Variant)

    Returns:
        variant (models.ProfileVariant)
    """

    chrom = variant.CHROM
    if chrom.startswith(('chr', 'CHR', 'Chr')):
        chrom = chrom[3:]

    pos = int(variant.POS)

    variant_id = get_variant_id(variant)

    ref = variant.REF
    alt = variant.ALT[0]

    maf = get_maf(variant)

    profile_variant = ProfileVariant(
        variant_id=variant_id,
        chrom=chrom,
        pos=pos,
        ref=ref,
        alt=alt,
        maf=maf,
        id_column = variant.ID
    )

    return profile_variant



import logging
import click
import coloredlogs

from pprint import pprint as pp

import yaml

from mongomock import MongoClient as MockClient

from pymongo import uri_parser

from mongo_adapter import get_client
from mongo_adapter.exceptions import Error as DB_Error

from loqusdb.log import LEVELS, init_log
from loqusdb import __version__
from loqusdb.plugins import MongoAdapter
from loqusdb.constants import GRCH37, GRCH38

LOG_LEVELS = ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']

LOG = logging.getLogger(__name__)

@click.group()
@click.option('-db', '--database',
                help="Defaults to 'loqusdb' if not specified",
)
@click.option('-u', '--username',
                type=str
)
@click.option('-p', '--password',
                type=str
)
@click.option('-a', '--authdb',
                type=str,
                help="If authentication should be done against another database than --database"
)
@click.option('-port', '--port',
                default=27017,
                show_default=True,
                help='Specify the port where to look for the mongo database.'
)
@click.option('-h', '--host',
                default='localhost',
                show_default=True,
                help='Specify the host where to look for the mongo database.'
)
@click.option('--uri',
                help='Specify a mongodb uri'
)
@click.option('-c', '--config',
                type=click.File('r'),
                help='Use a config with db information'
)
@click.option('-t', '--test',
                is_flag=True,
                help='Used for testing. This will use a mongomock database.'
)
@click.option('-g', '--genome-build',
              type=click.Choice([GRCH37, GRCH38]),
              help='Specify what genome build to use')
@click.option('-v', '--verbose', is_flag=True)
@click.version_option(__version__)
@click.pass_context
def cli(ctx, database, username, password, authdb, port, host, uri, verbose, config, test, genome_build):
    """loqusdb: manage a local variant count database."""
    loglevel = "INFO"
    if verbose:
        loglevel = "DEBUG"
    coloredlogs.install(level=loglevel)
    LOG.info("Running loqusdb version %s", __version__)

    configs = {}
    if config:
        try:
            configs = yaml.safe_load(config)
        except yaml.YAMLError as err:
            LOG.warning(err)
            ctx.abort()

    uri = configs.get('uri') or uri
    if test:
        uri = "mongomock://"
    try:
        client = get_client(
            host=configs.get('host') or host,
            port=configs.get('port') or port,
            username=configs.get('username') or username,
            password=configs.get('password') or password,
            authdb=authdb or database or 'loqusdb',
            uri=uri,
        )
    except DB_Error as err:
        LOG.warning(err)
        ctx.abort()

    database = configs.get('db_name') or database

    if not database:
        database = 'loqusdb'
        if uri:
            uri_info = uri_parser.parse_uri(uri)
            database = uri_info.get('database')

    adapter = MongoAdapter(client, db_name=database)

    genome_build = genome_build or configs.get('genome_build') or GRCH37

    ctx.obj = {}
    ctx.obj['db'] = database
    ctx.obj['user'] = username
    ctx.obj['password'] = password
    ctx.obj['port'] = port
    ctx.obj['host'] = host
    ctx.obj['adapter'] = adapter
    ctx.obj['version'] = __version__
    ctx.obj['genome_build'] = genome_build



# -*- coding: utf-8 -*-
import os
import logging
import subprocess

from datetime import datetime

import click

from . import base_command

LOG = logging.getLogger(__name__)

@base_command.command('dump', short_help="Dump the database")
@click.option('-f' ,'--filename', 
                help='If custom named file is to be used',
                type=click.Path(exists=False)
)
@click.pass_context
def dump(ctx, filename):
    """Dump the database to a zipped file.
    
    Default filename is loqusdb.<todays date>.gz (e.g loqusdb.19971004.gz)
    """
    
    if not filename:
        filename = "loqusdb.{}.gz".format(datetime.now().strftime('%Y%m%d'))
    
    if os.path.isfile(filename):
        LOG.warning("File {} already exists. Please remove file or change name with '--filename'".format(filename))
        ctx.abort()
    
    call = ['mongodump', '--gzip', '--db', 'loqusdb', '--archive={}'.format(filename)]
    
    LOG.info('dumping database...')
    start_time = datetime.now()
    try:
        completed = subprocess.run(call, check=True)
    except subprocess.CalledProcessError as err:
        LOG.warning(err)
        LOG.info("Deleting dump..")
        os.path.remove(filename)
        ctx.abort()
    
    LOG.info('Database dumped succesfully')
    LOG.info('Time to dump database: {0}'.format(datetime.now()-start_time))
    
    


# -*- coding: utf-8 -*-
import logging
import click
from datetime import datetime

from loqusdb.exceptions import CaseError
from loqusdb.utils.delete import delete as delete_command
from loqusdb.utils.case import get_case
from . import base_command

LOG = logging.getLogger(__name__)

@base_command.command('delete', short_help="Delete the variants of a family")
@click.option('-f', '--family-file',
                    type=click.Path(exists=True),
                    metavar='<ped_file>'
)
@click.option('-t' ,'--family-type',
                type=click.Choice(['ped', 'alt', 'cmms', 'mip']),
                default='ped',
                help='If the analysis use one of the known setups, please specify which one.'
)
@click.option('-c' ,'--case-id',
                type=str,
                help='If a different case id than the one in ped file should be used'
)
@click.pass_context
def delete(ctx, family_file, family_type, case_id):
    """Delete the variants of a case."""
    if not (family_file or case_id):
        LOG.error("Please provide a family file")
        ctx.abort()

    adapter = ctx.obj['adapter']

    # Get a ped_parser.Family object from family file
    family = None
    family_id = None
    if family_file:
        with open(family_file, 'r') as family_lines:
            family = get_case(
                family_lines=family_lines,
                family_type=family_type
            )
            family_id = family.family_id

    # There has to be a case_id or a family at this stage.
    case_id = case_id or family_id

    if not case_id:
        LOG.warning("Please provide a case id")
        ctx.abort()

    existing_case = adapter.case({'case_id': case_id})
    if not existing_case:
        LOG.warning("Case %s does not exist in database" %case_id)
        return

    genome_build = ctx.obj['genome_build']
    start_deleting = datetime.now()
    try:
        delete_command(
            adapter=adapter,
            case_obj=existing_case,
            genome_build=genome_build
        )
    except (CaseError, IOError) as error:
        LOG.warning(error)
        ctx.abort()





class ProfileVariant(dict):
    """Represent a profile variant

    These variants will be used to profile samples, based on a samples genotype
    at these positions.
    """

    def __init__(self, variant_id, chrom, pos, ref, alt, maf, id_column=None):
        super(ProfileVariant, self).__init__(
            _id=variant_id,
            chrom=chrom,
            pos=pos,
            ref=ref,
            alt=alt,
            maf=maf,
            id_column = id_column
        )



import pytest

from pprint import pprint as pp

from loqusdb.commands import base_command

from loqusdb.exceptions import ProfileError

from click.testing import CliRunner

def test_base_command():
    runner = CliRunner()
    result = runner.invoke(base_command, [])

    assert result.exit_code == 0


def test_load_command(vcf_path, ped_path, real_mongo_adapter, real_db_name):
    ## GIVEN a vcf_path a ped_path and a empty database
    assert sum([1 for case in real_mongo_adapter.cases()]) == 0
    runner = CliRunner()
    ## WHEN inserting a case via the CLI
    command = ['--database', real_db_name, 'load', '--variant-file', vcf_path, '-f', ped_path]
    result = runner.invoke(base_command, command)
    ## THEN assert that the cli exits without problems
    assert result.exit_code == 0
    ## THEN assert that the case was added
    assert sum([1 for case in real_mongo_adapter.cases()]) == 1

def test_load_command_no_ped_case_id(vcf_path, case_id, real_mongo_adapter, real_db_name):
    ## GIVEN a vcf_path a ped_path and a empty database
    assert real_mongo_adapter.case({'case_id':case_id}) is None
    runner = CliRunner()
    ## WHEN inserting a case via the CLI
    command = ['--database', real_db_name, 'load', '--variant-file', vcf_path, '-c', case_id]
    result = runner.invoke(base_command, command)
    ## THEN assert that the cli exits without problems
    assert result.exit_code == 0
    ## THEN assert that the case was added
    assert real_mongo_adapter.case({'case_id':case_id})

def test_load_command_no_ped(vcf_path, ped_path, real_mongo_adapter, real_db_name):
    ## GIVEN a vcf_path a ped_path and a empty database
    runner = CliRunner()
    ## WHEN inserting a case via the CLI without a ped file
    command = ['--database', real_db_name, 'load', '--variant-file', vcf_path]
    result = runner.invoke(base_command, command)

    ## THEN assert that the cli exits with code 1
    assert result.exit_code == 1

def test_load_command_check_profile(zipped_vcf_path, ped_path, real_mongo_adapter, real_db_name, profile_vcf_path):

    NEW_CASE_NAME = 'other_case'
    #Load profile variants
    runner = CliRunner()
    command = ['--database', real_db_name, 'profile', '--load',
        '--variant-file', profile_vcf_path]
    result = runner.invoke(base_command, command)

    #WHEN load command with --check-profile flag
    command = ['--database', real_db_name, 'load', '--variant-file', zipped_vcf_path, '-f',
                ped_path, '--check-profile', zipped_vcf_path]
    result = runner.invoke(base_command, command)
    ## THEN assert that the cli exits without problems
    assert result.exit_code == 0
    ## THEN assert if field 'profile is found in the individual documents'
    for case in real_mongo_adapter.cases():
        for individual in case['individuals']:
            assert individual.get('profile')

    ##WHEN load the same case (change case-id with -c option)
    command = ['--database', real_db_name, 'load', '--variant-file', zipped_vcf_path, '-f',
                ped_path, '--check-profile', zipped_vcf_path, '-c', 'other_case']
    result = runner.invoke(base_command, command)
    ## THEN assert that exit_code is not 0 and error is raised
    assert result.exit_code != 0
    assert type(result.exception) == ProfileError

    ##WHEN try to insert the same case without --check-profile flag (should work)
    command = ['--database', real_db_name, 'load', '--variant-file', zipped_vcf_path, '-f',
                ped_path, '-c', NEW_CASE_NAME]
    result = runner.invoke(base_command, command)
    ## THEN assert that exit_code is 0
    assert result.exit_code == 0
    ## THEN assert that no profile string is added to the individuals
    for individual in real_mongo_adapter.case({'case_id': NEW_CASE_NAME})['individuals']:
        assert not individual.get('profile')


def test_load_profile_command(profile_vcf_path, real_mongo_adapter, real_db_name):

    runner = CliRunner()

    command = ['--database', real_db_name, 'profile', '--load',
        '--variant-file', profile_vcf_path]
    result = runner.invoke(base_command, command)

    assert result.exit_code == 0
    assert len(list(real_mongo_adapter.profile_variants())) > 0

def test_delete_command_family_file(vcf_path, ped_path, real_mongo_adapter, real_db_name):
    ## GIVEN a vcf_path a ped_path and a empty database
    runner = CliRunner()
    assert sum([1 for case in real_mongo_adapter.cases()]) == 0

    ## WHEN inserting a case via the CLI without a ped file
    load_command = ['--database', real_db_name, 'load', '--variant-file', vcf_path, '-f', ped_path]
    result = runner.invoke(base_command, load_command)
    ## THEN assert that the case was added
    assert sum([1 for case in real_mongo_adapter.cases()]) == 1

    ## WHEN deleting the case a case via the CLI without a ped file
    delete_command = ['--database', real_db_name, 'delete', '-f', ped_path]
    result = runner.invoke(base_command, delete_command)

    ## THEN assert that the cli exits without problems
    assert result.exit_code == 0
    ## THEN assert that the case was deleted
    assert sum([1 for case in real_mongo_adapter.cases()]) == 0

def test_delete_command_case_id(vcf_path, case_id, real_mongo_adapter, real_db_name):
    ## GIVEN a vcf_path a ped_path and a empty database
    runner = CliRunner()
    assert real_mongo_adapter.case({'case_id': case_id}) is None

    ## WHEN inserting a case via the CLI without a ped file
    load_command = ['--database', real_db_name, 'load', '--variant-file', vcf_path, '-c', case_id]
    result = runner.invoke(base_command, load_command)
    ## THEN assert that the case was added
    assert isinstance(real_mongo_adapter.case({'case_id': case_id}), dict)

    ## WHEN deleting the case a case via the CLI without a ped file
    delete_command = ['--database', real_db_name, 'delete', '-c', case_id]
    result = runner.invoke(base_command, delete_command)

    ## THEN assert that the cli exits without problems
    assert result.exit_code == 0
    ## THEN assert that the case was deleted
    assert real_mongo_adapter.case({'case_id': case_id}) is None

def test_cases_command_case_id(vcf_path, case_id, real_mongo_adapter, real_db_name):
    ## GIVEN a vcf_path a ped_path and a empty database
    runner = CliRunner()
    assert real_mongo_adapter.case({'case_id': case_id}) is None

    ## WHEN inserting a case via the CLI without a ped file
    load_command = ['--database', real_db_name, 'load', '--variant-file', vcf_path, '-c', case_id]
    result = runner.invoke(base_command, load_command)
    ## THEN assert that the case was added
    assert isinstance(real_mongo_adapter.case({'case_id': case_id}), dict)

    ## WHEN searching for the case with CLI
    command = ['--database', real_db_name, 'cases', '--case-id', case_id]
    result = runner.invoke(base_command, command)

    ## THEN assert that the cli exits without problems
    assert result.exit_code == 0


def test_cases_command_non_existing(vcf_path, case_id, real_mongo_adapter, real_db_name):
    ## GIVEN a vcf_path a ped_path and a empty database
    runner = CliRunner()
    assert real_mongo_adapter.case({'case_id': case_id}) is None

    ## WHEN searching for a non existing case
    command = ['--database', real_db_name, 'cases', '--case-id', 'hello']
    result = runner.invoke(base_command, command)

    ## THEN assert that the exit is zero
    assert result.exit_code == 0
    # Exit code should be 0 even if there is no result returned



# def test_wipe_command(mongo_client, vcf_path, ped_path):
#     runner = CliRunner()
#
#     load_command = ['-c', 'mongomock://', '--database', 'test', 'load', vcf_path, '-f', ped_path]
#     runner.invoke(base_command, load_command)
#
#     wipe_command = ['-c', 'mongomock://', '--database', 'test', 'wipe', '--yes']
#     result = runner.invoke(base_command, wipe_command)
#
#     assert result.exit_code == 0



import click
import logging
import json

from loqusdb.utils.load import load_profile_variants
from loqusdb.utils.profiling import (update_profiles,
                                     profile_stats,
                                     check_duplicates,
                                     get_profiles)
from loqusdb.resources import MAF_PATH

from . import base_command

LOG = logging.getLogger(__name__)


def validate_profile_threshold(ctx, param, value):
    if not (0 <= value <= 1):
        raise ValueError('threshold must be between 0-1')

    return value

@base_command.command('profile', short_help = "Loads variants to be used in profiling")
@click.option('-l', '--load',
    is_flag=True,
    help = "Load variants that should be used for profiling")
@click.option('-v', '--variant-file',
    type=click.Path(exists=True),
    help="Variants from file to be used in profiling. To be used in combination with the --load flag")
@click.option('--update',
    is_flag=True,
    help = "updates the profiles of all the sample in the database")
@click.option('--stats',
    is_flag=True,
    help="Checks some statistics of the profiles in the database")
@click.option('--profile-threshold',
    type=float,
    default=0.9,
    callback=validate_profile_threshold,
    help="Used with --stats option to determine the number of matching profiles with a similarity greater than given threshold")
@click.option('--check-vcf',
    type=click.Path(exists=True),
    help="A vcf for a case. The profile from this vcf will be checked against the profiles in the database")
@click.pass_context
def load_profile(ctx, load, variant_file, update, stats, profile_threshold, check_vcf):

    """
        Command for profiling of samples. User may upload variants used in profiling
        from a vcf, update the profiles for all samples, and get some stats
        from the profiles in the database.

        Profiling is used to monitor duplicates in the database. The profile is
        based on the variants in the 'profile_variant' collection, assessing
        the genotypes for each sample at the position of these variants.
    """

    adapter = ctx.obj['adapter']

    LOG.info("Running loqusdb profile")

    if check_vcf:
        LOG.info(f"Check if profile in {check_vcf} has match in database")
        vcf_file = check_vcf
        profiles = get_profiles(adapter, vcf_file)
        duplicate = check_duplicates(adapter, profiles, profile_threshold)

        if duplicate is not None:
            duplicate = json.dumps(duplicate)
            click.echo(duplicate)
        else:
            LOG.info("No duplicates found in the database")

    if load:
        genome_build = ctx.obj['genome_build']
        vcf_path = MAF_PATH[genome_build]
        if variant_file is not None:
            vcf_path = variant_file
        LOG.info(f"Loads variants in {vcf_path} to be used in profiling")
        load_profile_variants(adapter, vcf_path)

    if update:
        LOG.info("Updates profiles in database")
        update_profiles(adapter)

    if stats:
        LOG.info("Prints profile stats")
        distance_dict = profile_stats(adapter, threshold=profile_threshold)
        click.echo(table_from_dict(distance_dict))

def table_from_dict(dictionary):


    table_str = "Distances within ranges:\n"

    for key, value in dictionary.items():

        table_str += f"{key:15} | {value}\n"

    return table_str



# -*- coding: utf-8 -*-
import os
import logging
import subprocess

from datetime import datetime

import click

from . import base_command
from loqusdb.resources import background_path

LOG = logging.getLogger(__name__)

@base_command.command('restore', short_help="Restore database from dump")
@click.option('-f' ,'--filename', 
                help='If custom named file is to be used',
                type=click.Path(exists=True),
)
@click.pass_context
def restore(ctx, filename):
    """Restore the database from a zipped file.
    
    Default is to restore from db dump in loqusdb/resources/
    """
    filename = filename or background_path
    if not os.path.isfile(filename):
        LOG.warning("File {} does not exist. Please point to a valid file".format(filename))
        ctx.abort()
    
    call = ['mongorestore', '--gzip', '--db', 'loqusdb', '--archive={}'.format(filename)]
    
    LOG.info('Restoring database from %s', filename)
    start_time = datetime.now()
    try:
        completed = subprocess.run(call, check=True)
    except subprocess.CalledProcessError as err:
        LOG.warning(err)
        ctx.abort()
    
    LOG.info('Database restored succesfully')
    LOG.info('Time to restore database: {0}'.format(datetime.now()-start_time))
    
    


from loqusdb.build_models import build_variant
from loqusdb.build_models.variant import get_coords

def test_build_het_variant(het_variant, case_obj):
    variant_obj = build_variant(
        variant = het_variant,
        case_obj = case_obj
    )
    assert variant_obj['chrom'] == het_variant.CHROM
    assert variant_obj['homozygote'] == 0
    assert variant_obj['hemizygote'] == 0


def test_get_coords_for_BND(bnd_variant):
    coords = get_coords(bnd_variant)
    assert coords['pos'] == coords['end']
    assert coords['sv_length'] == float('inf')
    assert coords['sv_type'] == 'BND'
